# Copyright 2021-2022 Timo Röhling <roehling@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
package Debian::Debhelper::Buildsystem::ros;

use strict;
use Debian::Debhelper::Dh_Lib qw(doit);
use base 'Debian::Debhelper::Buildsystem';

sub DESCRIPTION {
    "ROS package builder"
}

sub check_auto_buildable {
    my $this = shift;
    return doit('rosbuilder', '--detect', '--sourcedir', $this->get_sourcedir());
}

sub new {
    my $class = shift;
    my $this = $class->SUPER::new(@_);
    $this->prefer_out_of_source_building();
    return $this;
}

sub configure {
    return 1
}

sub build {
    return 1
}

sub test {
    return 1
}

sub install {
    my $this = shift;
    my $destdir = shift;
    return doit('rosbuilder', '--sourcedir', $this->get_sourcedir(), '--builddir', $this->get_builddir(), '--destdir', $destdir, @_);
}

sub clean {
    my $this = shift;
    return doit('rm', '-rf', $this->get_builddir()) && doit('rm', '-rf', '.pybuild');
}

1
